﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using NLog;
using System.IO;
using System.Windows.Threading;
using System.Threading;

namespace AlarmReader
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        #region Constants

        public const int ReceivePort = 162;
        public enum RpResult { RP_SUCCESS, RP_FAIL, RP_FATAL };

        #endregion

        #region Private Members

        SnmpTlv _snmpPacketTlv;					// Current snmp packet for logging.

        private SocketSupport mSocketSupport;
        protected string _receiveResult;
        private static Random m_RandomNumberGenerator = null;

        private static Logger mLogger = LogManager.GetCurrentClassLogger();

        #endregion

        public MainWindow()
        {
            InitializeComponent();

            mSocketSupport = new SocketSupport();
            mSocketSupport.ReceivePort = ReceivePort;

            SocketSupport.ReceiveDelegateH temp = new SocketSupport.ReceiveDelegateH(ReceiveCallbackSnmp);
            SocketSupport.ReceiveDelegateE tempErr = new SocketSupport.ReceiveDelegateE(ErrorStop);
            mSocketSupport.SetCallbackH(temp, tempErr);

        }

        /// <summary>
        /// Append the message to the result box on screen
        /// </summary>
        /// <param name="aMessage"></param>
        public void DisplayResultMessage(string aMessage)
        {
            ResultBox.Dispatcher.Invoke(new Action(
                    delegate()
                    {
                        mLogger.Info("DisplayResultMessage:1: writing '"+aMessage+"'");
                        ResultBox.Text+= aMessage+ "\r\n";
                    }));
        }

        #region SNMP Packet receipt and stopping

        /// <summary>
        /// This stops the receiver due to error.
        /// </summary>
        private void ErrorStop()
        {
            try
            {
                mLogger.Error(""); // blank before new event
                mLogger.Error("MainWindow.ErrorStop: start");

                // Shut ourselves down
                mSocketSupport.IsShuttingDown = true;   // Stop any socket restarts

                try
                {
                    if (mSocketSupport.ReceiveSocket != null)
                    {
                        mSocketSupport.ReceiveSocket.Close();
                        mSocketSupport.ReceiveSocket = null;
                    }
                }
                catch { }

                mSocketSupport.IsRunning = false;

            }
            catch (Exception ex)
            {
                mLogger.Error("ErrorStop: got ex: "+ex);
            }

        }

        /// <summary>
        /// This is called each time a UDP packet is received on our SNMP receive port.
        /// We process it and set up for the next receive.
        /// As part of that processing, many lines are logged.
        /// </summary>
        /// <param name="result">Contains socket and other info.</param>
        private void ReceiveCallbackSnmp(IAsyncResult result)
        {
            try
            {
                mLogger.Trace("ReceiveCallbackSnmp: start");
                _receiveResult = "Receive result not initialized!";

                // Force blank lines before and after
                StringBuilder sb = new StringBuilder();
                sb.AppendLine("");
                sb.AppendLine("====================== Incoming alarm ======================");
                sb.AppendLine("");
                SnmpTlv.WriteAlarmLog(sb.ToString());

                // Read the message in from the SNMP port, peforming logging if needed.
                if (ReadMessageSummary(result) == false)
                    return;
            }
            catch (Exception ex)
            {
                mLogger.Error("ReceiveCallbackSnmp: Got ex:" + ex);
            }
        }

        #endregion

        /// <summary>
        /// Return a random integer between and including the 2 given values.
        /// </summary>
        /// <param name="minimum">minimum value of range</param>
        /// <param name="maximum">maximum value of range</param>
        /// <returns></returns>
        static public int RandomInt(int minimum, int maximum)
        {
            if (m_RandomNumberGenerator == null)
            {
                m_RandomNumberGenerator = new Random();
            }

            // quirk of Next method is that it won't give you the highest value back
            return m_RandomNumberGenerator.Next(minimum, maximum + 1);
        }

        #region ReadMessage routines

        /// <summary>
        /// Read in the UDP packet and log its content
        /// </summary>
        /// <param name="result">Socket info to complete a read</param>
        /// <returns>True on success, false on fail.</returns>
        private bool ReadMessageSummary(IAsyncResult result)
        {
            // Read in the snmp message and convert it to a list of strings
            RpResult rc = ReadMessage(result);

            if (rc == RpResult.RP_FAIL)
            {
                DisplayResultMessage("ReadMessage non-fatal error.");
                return false;
            }
            else if (rc == RpResult.RP_FATAL)
            {
                DisplayResultMessage("ReadMessage fatal error - see log.");
                mLogger.Error("ReceiveCallback: ReadMessage fatal error");

                return false;
            }

            // Success
            DisplayResultMessage("Received alarm - details in log.");
            return true;
        }

        /// <summary>
        /// Read in the UDP packet and return it as a list of strings
        /// </summary>
        /// <param name="result">Socket info to complete a read</param>
        /// <returns>RpResult of SUCCESS, FAIL, or FATAL</returns>
        private RpResult ReadMessage(IAsyncResult result)
        {
            // Note entry
            mLogger.Trace("ReadMessage: start");

            // Receive our packet.
            if (mSocketSupport.CompleteUdpRead(result) < 0)  // Failed to read packet
            {
                _receiveResult = "Failed to complete UDP read!";
                mLogger.Error("ReadMessage: " + _receiveResult);

                if (mSocketSupport.IsShuttingDown == true)
                    return RpResult.RP_FAIL;
                else
                    return RpResult.RP_FATAL;
            }

            if (mSocketSupport.InBufferLength < 1)
            {
                _receiveResult = "Discarded empty packet";
                mLogger.Info("ReadParseMessage: " + _receiveResult);
                return RpResult.RP_FAIL;
            }

            mLogger.Debug("ReadMessage: completed udp read of at least 1 byte");

            // The _inbuffer length is fixed at 65536, we need the array to be of actual buffer length.
            byte[] tempBuf = SnmpTlv.CopySection(mSocketSupport.InBuffer, 0, mSocketSupport.InBufferLength);
            mSocketSupport.InBuffer = tempBuf;

            if (mSocketSupport.InBuffer.Length != mSocketSupport.InBufferLength)
            {
                mLogger.Info("ReadMessage: _inbuffer array length(" + mSocketSupport.InBuffer.Length +
                    "!= _inbufferLength(" + mSocketSupport.InBufferLength + ")");
            }

            // Create the TLV of packet, for later logging if needed
            try
            {
                _snmpPacketTlv = new SnmpTlv(mSocketSupport.InBuffer, false);

                _snmpPacketTlv.SnmpComment();
                _snmpPacketTlv.AlarmLogit(0);
            }
            catch (Exception ex)
            {
                mLogger.Error("SnmpManager.ReadMessage: Couldn't do SnmpTlv on packet, ex=" + ex);
                LogBytes(LogLevel.Info, mSocketSupport.InBuffer, mSocketSupport.InBuffer.Length);
                SnmpTlv.AlarmLogBytes(mSocketSupport.InBuffer, mSocketSupport.InBufferLength, 0);  // alarmYYYY-MM-DD.log
            }

            // Operation completed
            _receiveResult = "Message read successfully";
            return RpResult.RP_SUCCESS;
        }


        #endregion

        #region LogBytes

        /// <summary>
        /// Log the given buffer, displaying printable characters. No indentation
        /// </summary>
        public static void LogBytes(LogLevel lvl, byte[] buffer, int len)
        {
            LogBytes(lvl, buffer, len, 0);
        }

        /// <summary>
        /// Log the given buffer, displaying printable characters
        /// </summary>
        public static void LogBytes(LogLevel lvl, byte[] buffer, int len, int indent)
        {
            // Set num of spaces for indentation
            string spc = "";
            if (indent > 0)
                spc = "                                                                                                                                    ".Substring(0, indent);

            if (buffer == null || len < 0)
            {
                mLogger.Error(spc + "LogBytes: invalid buffer or len argument");
                return;
            }

            mLogger.Log(lvl, spc + "LogBytes: buffer length=" + len);

            int i;
            StringBuilder hexPart = new StringBuilder(48);
            StringBuilder charPart = new StringBuilder(10);

            for (i = 0; i < len; i++)
            {
                if (i % 10 == 0 && i > 0)    // restart for new line
                {
                    mLogger.Log(lvl, spc + String.Format("    {0,-48}    {1}", hexPart, charPart));
                    hexPart.Length = 0;
                    charPart.Length = 0;
                }

                hexPart.Append(String.Format("  {0:x2}", buffer[i]));

                if (buffer[i] >= 32 && buffer[i] < 127)
                    charPart.Append((char)buffer[i]);
                else
                    charPart.Append(".");
            }

            if (hexPart.Length > 0)   // add ending newline
                mLogger.Log(lvl, spc + String.Format("    {0,-48}    {1}", hexPart, charPart));
        }

        #endregion

        #region Button Event Handlers

        /// <summary>
        /// Start monitoring the port
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void StartButton_Click(object sender, RoutedEventArgs e)
        {
            bool rc;

            mLogger.Info("StartButton_Click:1: start monitor");
            DisplayResultMessage("Starting monitor of port " + ReceivePort + ".");

            // Open the socket, setting up our callback listener
            rc = mSocketSupport.RestartSocket();

        }

        /// <summary>
        /// Stop monitoring the port.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void EndButton_Click(object sender, RoutedEventArgs e)
        {
            mLogger.Info("EndButton_Click:1: stop monitor");
            DisplayResultMessage("Stopping monitor of port " + ReceivePort + ".");

            mSocketSupport.Stop();
        }

        /// <summary>
        /// Send a random test alarm.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void TestButton_Click(object sender, RoutedEventArgs e)
        {
            int dur;

            // Get the number of alarms from text box
            int numAlarms = 1;
            mLogger.Info("TestButton_Click:1: numAlarms=" + numAlarms);

            // Send requested number of random alarms
            int succCnt = 0;
            int failCnt = 0;

            // WPF queues up the draws until we are finished, so do redraw event now.
            DoEvents();

            for (int i = 0; i < numAlarms; i++)
            {
                int type = RandomInt(1, TestAlarm.HIGHEST_SNMP_ALARM_TYPE);
                dur = TestAlarm.Send1SnmpAlarm(type, "localhost", MainWindow.ReceivePort);

                // Act on result
                if (dur >= 0)
                {
                    succCnt++;
                }
                else
                {
                    failCnt++;
                }
            }

            DisplayResultMessage("Sent " + succCnt + " snmp alarms successfully, " + failCnt + " failed.");
        }
        
        /// <summary>
        /// Carry out any waiting windows events while a routine is executing.
        /// </summary>
        private static void DoEvents()
        {
            Application.Current.Dispatcher.Invoke(DispatcherPriority.Background, new ThreadStart(delegate { }));
        }

        #endregion

    }
}
