﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NLog;
using SNMPSharp;
using System.IO;

namespace AlarmReader
{
    /// <summary>
    /// Type,List,Value trio that SNMP uses extensively
    /// </summary>
    public class SnmpTlv
    {
        #region Constants

        // Byte list types
        public enum BYTE_LIST_TYPE { OCTSTR, IPADDR, OID, TIMETICKS };
        public static readonly byte HIGH_BIT = (byte)0x80;

        // TimeTicks is supposed to be a Uint32 (uint)

        const int ASN_INT = 0x02;       // ASN.1 integer
        const int ASN_OCTSTR = 0x04;    // ASN.1 octet string
        const int ASN_NULL = 0x05;      // null
        const int ASN_OBJID = 0x06;     // Object identifier (oid)
        const int ASN_SEQUENCE = 0x30;  // Sequence code

        const int ASN_IPADDR = 0x40;    // An IPv4 address
        const int ASN_COUNTER = 0x41;
        const int ASN_GAUGE = 0x42;
        const int ASN_TIMETICKS = 0x43; // Time tick value
        const int ASN_OPAQUE = 0x44;
        const int ASN_NSAP_ADDRESS = 0x45;
        const int ASN_COUNTER64 = 0x46;
        const int ASN_UINTEGER32 = 0x47;

        const int ASN_GETREQPDU = 0xa0;    // Get request PDU
        const int ASN_GETRESPPDU = 0xa2;    // Get response PDU
        const int ASN_TRAPPDU = 0xa4;    // Trap PDU
        const int ASN_V2GETBULKPDU = 0xa5;
        const int ASN_V2INFORMPDU = 0xa6;
        const int ASN_V2TRAPPDU = 0xa7; // Snmp v2's trap

        #endregion

        byte _Type;      // ASN.1 type for this item
        int _Length;       // Length of this item
        string _Comment;    // Note on meaning of this field

        // Simple types will only have _Packet, Sequence will have child(ren).
        List<SnmpTlv> _ChildList;   // Underlying TLV's
        byte[] _Packet;     // Underlying packet octets
        byte[] _LengthBytes;  // Actual bytes making up length part

        private static Logger mLogger = LogManager.GetCurrentClassLogger();

        /// <summary>
        /// Constructor with packet of bytes to decode. inSequence=true means this is an internal sequence,
        /// and lengths don't have to match. This throws Exception on error.
        /// </summary>
        /// <param name="aPacket"></param>
        /// <param name="inSequence"></param>
        public SnmpTlv(byte[] aPacket, bool inSequence)
        {
            // Get our type and length, assume each is one byte for now.
            _Type = aPacket[0];

            // Determine length. BER encoding uses a high 1-bit to indicate the 7 bits are actually a count
            // of the following bytes. Ex: 0x82, 0x01, 0x7b . 82=Following 2 bytes. Len=0x017b, or 379 bytes.
            int lenCount = 1;

            if ((aPacket[1] & 0x80) != 0)  // Byte count method.
            {
                lenCount = (aPacket[1] & 0x7f) + 1;  // byte counter itself

                if (lenCount == 2)
                    _Length = aPacket[2];
                else if (lenCount == 3)
                    _Length = aPacket[2] * 256 + aPacket[3];
                else
                    _Length = aPacket[2] * 65536 + aPacket[3] * 256 + aPacket[4];
            }
            else // Ordinary
                _Length = aPacket[1];
            _LengthBytes = CopySection(aPacket, 1, lenCount);

            mLogger.Trace("SnmpTlv:1: aPacket.Length=" + aPacket.Length + ", hdrLen=" +
                GetHeaderLength() + ", _Length=" + _Length +
                " . LengthBytes:");

             _Comment = "";


            // ChildList will stay empty if this is not a sequence
            _ChildList = new List<SnmpTlv>();

            // Need to adjust if length is wrong 
            if (GetHeaderLength() + _Length > aPacket.Length || (inSequence == false && GetHeaderLength() + _Length != aPacket.Length))
            {
                throw new Exception("SnmpTlv: TLV length of " + _Length + GetHeaderLength() + "differs from packet length " +
                    aPacket.Length);
            }

            // Set our packet to just the bytes we are using.
            _Packet = CopySection(aPacket, GetHeaderLength(), _Length);

            if (_Packet == null)
            {
                throw new Exception("SnmpTlv: Unable to create packet of length " + _Length);
            }

            mLogger.Trace("SnmpTlv:2: Type=" + _Type + "(SEQ="+ASN_SEQUENCE+"), Length=" + _Length);

            if (_Type == ASN_SEQUENCE || _Type == ASN_TRAPPDU || _Type == ASN_V2TRAPPDU)
            {
                UnpackSequence();
            }
        }

        /// <summary>
        /// Return the length of the header section of the TLV in bytes.
        /// </summary>
        /// <returns></returns>
        public int GetHeaderLength()
        {
            return 1 /* Type */ + _LengthBytes.Length;
        }

        /// <summary>
        /// Since CopyTo doesn't seem to have a way to copy a part inside the array, this returns a
        /// new array that contains the elements from aStartIdx for aLength bytes.
        /// Returns null on error, or the new array on success.
        /// </summary>
        /// <param name="arg"></param>
        /// <param name="aStartIdx"></param>
        /// <param name="aLength"></param>
        /// <returns></returns>
        static public byte[] CopySection(byte[] arg, int aStartIdx, int aLength)
        {
            if (arg == null || aStartIdx + aLength > arg.Length)
            {
                mLogger.Error("SnmpTlv.CopySection: invalid args");
                return null;
            }

            byte[] newArray = new byte[aLength];
            int fromIdx, toIdx = 0;

            for (fromIdx = aStartIdx; toIdx < aLength; fromIdx++)
                newArray[toIdx++] = arg[fromIdx];

            return newArray;
        }

        /// <summary>
        /// Unpack a series of TLV's.
        /// </summary>
        /// <param name="aStartIdx"></param>
        private void UnpackSequence()
        {
            int curIdx = 0;

            mLogger.Debug("SnmpTlv.UnpackSequence: Start, Length=" + _Length);

            while (curIdx < _Length)
            {
                // Adjust to what's left
                byte[] subPacket = CopySection(_Packet, curIdx, _Length-curIdx);

                // One or more TLV's in a row
                SnmpTlv child = new SnmpTlv(subPacket, true);

                curIdx += child._Length + child.GetHeaderLength();
                _ChildList.Add(child);
            }

            mLogger.Trace("SnmpTlv.UnpackSequence: returning");
        }

        #region Logit to debug log

        public void Logit(LogLevel lvl, int indent)
        {
            
            switch (_Type)
            {
                case ASN_INT:
                    IntLogit(lvl, indent); break;
                case ASN_OCTSTR:
                    OctStrLogit(lvl, indent); break;
                case ASN_SEQUENCE:
                case ASN_TRAPPDU:
                case ASN_V2TRAPPDU:
                    SequenceLogit(lvl, indent); break;
                default:
                    OctStrLogit(lvl, indent); break;
            }
        }

        /// <summary>
        /// Write a debug line for an integer
        /// </summary>
        /// <param name="lvl"></param>
        /// <param name="indent"></param>
        public void IntLogit(LogLevel lvl, int indent)
        {
            string spc = SpaceString(indent);
            string line;

            // Format and writ ethe line to the log
            line = spc + TypeString() + ", " + LengthString() + 
                String.Format("  Value: 0x{0:x2} ({0} dec)", _Packet[0]);

            WriteHdrLine(line, lvl);            
        }

        /// <summary>
        /// Log the contents of a sequence, such as the packet as a whole or an individual Trap PDU.
        /// </summary>
        /// <param name="lvl"></param>
        /// <param name="indent"></param>
        public void SequenceLogit(LogLevel lvl, int indent)
        {
            string spc = SpaceString(indent);
            string line;
            int i;

            // Format and write header line to log
            line = spc + TypeString() + ", " + LengthString();
            WriteHdrLine(line, lvl);

            for (i = 0; i < _ChildList.Count; i++)
            {
                _ChildList[i].Logit(lvl, indent + 4);
            }

            mLogger.Info(spc+"END", lvl);
        }

        public void OctStrLogit(LogLevel lvl, int indent)
        {
            string spc = SpaceString(indent);
            string line;

            // Header line
            line = spc + TypeString() + ", " + LengthString();
            WriteHdrLine(line, lvl);

            // Log according to type
            switch (_Type)
            {
                case ASN_OBJID: ByteStringLogit(_Packet, lvl, indent + 4, BYTE_LIST_TYPE.OID); break;
                case ASN_IPADDR: ByteStringLogit(_Packet, lvl, indent + 4, BYTE_LIST_TYPE.IPADDR); break;
                case ASN_TIMETICKS: ByteStringLogit(_Packet, lvl, indent + 4, BYTE_LIST_TYPE.TIMETICKS); break;
                default: ByteStringLogit(_Packet, lvl, indent + 4, BYTE_LIST_TYPE.OCTSTR); break;
            }
        }

        /// <summary>
        /// Write requested line to log, followed by any comments.
        /// </summary>
        /// <param name="line"></param>
        /// <param name="lvl"></param>
        public void WriteHdrLine(string line, LogLevel lvl)
        {
            if (_Comment.Length > 0)
                line +=  "    // "+_Comment;

            mLogger.Info(line, lvl);
        }

        #endregion

        #region Logit to alarm log

        public void AlarmLogit(int indent)
        {
            switch (_Type)
            {
                case ASN_INT:
                    AlarmIntLogit(indent); break;
                case ASN_OCTSTR:
                    AlarmOctStrLogit(indent); break;
                case ASN_SEQUENCE:
                case ASN_TRAPPDU:
                case ASN_V2TRAPPDU:
                    AlarmSequenceLogit(indent); break;
                default:
                    AlarmOctStrLogit(indent); break;
            }
        }

        /// <summary>
        /// Write a debug line for an integer
        /// </summary>
        /// <param name="lvl"></param>
        /// <param name="indent"></param>
        public void AlarmIntLogit(int indent)
        {
            string spc = SpaceString(indent);
            string line;

            // Format and writ ethe line to the log
            line = spc + TypeString() + ", " + LengthString() +
                String.Format("  Value: 0x{0:x2} ({0} dec)", _Packet[0]);

            AlarmWriteHdrLine(line);
        }

        /// <summary>
        /// Log the contents of a sequence, such as the packet as a whole or an individual Trap PDU.
        /// </summary>
        /// <param name="lvl"></param>
        /// <param name="indent"></param>
        public void AlarmSequenceLogit(int indent)
        {
            string spc = SpaceString(indent);
            string line;
            int i;

            // Format and write header line to log
            line = spc + TypeString() + ", " + LengthString();
            AlarmWriteHdrLine(line);

            for (i = 0; i < _ChildList.Count; i++)
            {
                _ChildList[i].AlarmLogit(indent + 4);
            }

            WriteAlarmLog(spc + "END");
        }

        public void AlarmOctStrLogit(int indent)
        {
            string spc = SpaceString(indent);
            string line;

            // Header line
            line = spc + TypeString() + ", " + LengthString();
            AlarmWriteHdrLine(line);

            // Log according to type
            switch (_Type)
            {
                case ASN_OBJID: AlarmByteStringLogit(_Packet, indent + 4, BYTE_LIST_TYPE.OID); break;
                case ASN_IPADDR: AlarmByteStringLogit(_Packet, indent + 4, BYTE_LIST_TYPE.IPADDR); break;
                case ASN_TIMETICKS: AlarmByteStringLogit(_Packet, indent + 4, BYTE_LIST_TYPE.TIMETICKS); break;
                default: AlarmByteStringLogit(_Packet, indent + 4, BYTE_LIST_TYPE.OCTSTR); break;
            }
        }

        /// <summary>
        /// Write requested line to log, followed by any comments.
        /// </summary>
        /// <param name="line"></param>
        /// <param name="lvl"></param>
        public void AlarmWriteHdrLine(string line)
        {
            if (_Comment.Length > 0)
                line += "    // " + _Comment;

            WriteAlarmLog(line);
        }
        
        /// <summary>
        /// Log the given buffer, displaying printable characters
        /// </summary>
        public static void AlarmByteStringLogit(byte[] buffer, int indent, BYTE_LIST_TYPE blt)
        {
            int len = buffer.Length;

            if (buffer == null || buffer.Length < 1)
            {
                mLogger.Debug("AlarmByteStringLogit: Null or empty buffer, returning");
                return;
            }

            // Set num of spaces for indentation
            string spc = SpaceString(indent);

            int i;
            StringBuilder hexPart = new StringBuilder(48);
            StringBuilder charPart = new StringBuilder(10);

            // For OID and IPADDR, charPart needs to be built up front
            if (blt == BYTE_LIST_TYPE.OID)
            {
                charPart.Append(OidString(buffer));
            }
            else if (blt == BYTE_LIST_TYPE.IPADDR)
            {
                for (i = 0; i < len; i++)
                {
                    if (i != 0)
                        charPart.Append(".");
                    charPart.Append(buffer[i]);
                }
            }
            else if (blt == BYTE_LIST_TYPE.TIMETICKS)
            {
                charPart.Append(TimeTickString(buffer));
            }

            for (i = 0; i < len; i++)
            {
                if (i % 10 == 0 && i > 0)    // restart for new line
                {
                    WriteAlarmLog(spc + String.Format("{0,-48}    {1}", hexPart, charPart));
                    hexPart.Length = 0;
                    charPart.Length = 0;
                    if (blt == BYTE_LIST_TYPE.IPADDR || blt == BYTE_LIST_TYPE.OID)
                        charPart.Clear();   // only want on first line
                }

                hexPart.Append(String.Format("  {0:x2}", buffer[i]));

                if (!(blt == BYTE_LIST_TYPE.IPADDR || blt == BYTE_LIST_TYPE.OID || blt == BYTE_LIST_TYPE.TIMETICKS))
                {
                    if (buffer[i] >= 32 && buffer[i] < 127)
                        charPart.Append((char)buffer[i]);
                    else
                        charPart.Append(".");
                }
            }

            if (hexPart.Length > 0)   // add ending newline
                WriteAlarmLog(spc + String.Format("{0,-48}    {1}", hexPart, charPart));
        }

        #endregion

        #region Alarm Logging
        
        /// <summary>
        /// Log the given buffer, displaying printable characters
        /// </summary>
        public static void AlarmLogBytes(byte[] buffer, int len, int indent)
        {
            StringBuilder sb = new StringBuilder();

            // Set num of spaces for indentation
            string spc = "";
            if (indent > 0)
                spc = "                                                                                                                                    ".Substring(0, indent);

            if (buffer == null || len < 0)
            {
                mLogger.Info(spc + "AlarmLogBytes: invalid buffer or len argument");
                return;
            }

            sb.AppendLine(spc + "AlarmLogBytes: buffer length=" + len);

            int i;
            StringBuilder hexPart = new StringBuilder(48);
            StringBuilder charPart = new StringBuilder(10);

            for (i = 0; i < len; i++)
            {
                if (i % 10 == 0 && i > 0)    // restart for new line
                {
                    sb.AppendLine(spc + String.Format("    {0,-48}    {1}", hexPart, charPart));
                    hexPart.Length = 0;
                    charPart.Length = 0;
                }

                hexPart.Append(String.Format("  {0:x2}", buffer[i]));

                if (buffer[i] >= 32 && buffer[i] < 127)
                    charPart.Append((char)buffer[i]);
                else
                    charPart.Append(".");
            }

            if (hexPart.Length > 0)   // add ending newline
                sb.AppendLine(spc + String.Format("    {0,-48}    {1}", hexPart, charPart));

            // Write completed string
            SnmpTlv.WriteAlarmLog(sb.ToString());
        }

        /// <summary>
        /// Write 1 line to the alarm log (open,write,close)
        /// </summary>
        /// <param name="aLine"></param>
        public static void WriteAlarmLog(string aLine)
        {
            const string DATE_FORMAT = "yyyy-MM-dd";
            string fPath = "logs\\alarm" + DateTime.Now.ToString(DATE_FORMAT) + ".log";

            try
            {
                TextWriter tw = new StreamWriter(fPath, true);
                tw.WriteLine(aLine);
                tw.Close();
            }
            catch (Exception ex)
            {
                mLogger.Info("WriteAlarmLog: got ex: " + ex);
            }
        }

        #endregion

        #region Item to String routines

        /// <summary>
        /// Return a string of blanks of appropriate length
        /// </summary>
        /// <param name="len"></param>
        /// <returns></returns>
        static public string SpaceString(int len)
        {
            string spc = "                                                                          " +
                "                                                                          " +
                "                                                                          " ;

            if (len == 0)
                return "";
            else
                return spc.Substring(0, len);
        }

        public string TypeString()
        {
            return "Type: " + HexByteString(_Type) + " (" + AsnToString(_Type) + ")";
        }

        public string LengthString()
        {
            if (_Length > 255)  // too big for a byte
                return "Length: " + String.Format("0x{0:x4} ({0} dec)", _Length);
            else
                return "Length: " + HexByteString((byte)(_Length & 0xff)) + " (" + _Length + " dec)";
        }

        public static string HexByteString(byte value)
        {
            return String.Format("0x{0:x2}", value);
        }

        public static string TimeTickString(byte[] buffer)
        {
            uint value =0, adjust;
            byte shift=0;
            int i;

            // Timeticks is supposed to be a 4 byte Uint32 (uint), however I've seen snmp records with longer.
            // We need to make sure there aren't non-zeroes before the last 4 bytes. Also need to handle less
            // than 4 bytes.
            for (i=buffer.Length-1; i>=0; i--)
            {
                if (shift <= 24)
                {
                    adjust = (uint)(buffer[i] << shift);
                    value |= adjust;
                    shift += 8;
                }
                else if (buffer[i] > 0)
                    return "Time value too big!";
            }

            TimeTicks tt = null;

            try
            {
                tt = new TimeTicks(value);
            }
            catch (Exception ex)
            {
                mLogger.Error("TimeTickString: Converting buffer len " + buffer.Length + " got ex: " + ex);
                return "Unable to convert value";
            }

            return tt.ToString();
        }

        public static string GenTrapString(int value)
        {
            switch (value)
            {
                case 0: return "Cold Start";
                case 1: return "Warm Start";
                case 2: return "Link Down";
                case 3: return "Link Up";
                case 4: return "Auth Failure";
                case 5: return "Neighbor Loss";
                case 6: return "See Specific trap";
                default: return "Unknown";
            }
        }

        /// <summary>
        /// Convert the OID component to it's equivalent string. This is based on the
        /// decode() method in the SNMP library.
        /// </summary>
        /// <param name="buffer"></param>
        /// <returns></returns>
        public static string OidString(byte[] buffer)
        {
            int offset = 0;
            int headerLength = buffer.Length;
            StringBuilder resultStr = new StringBuilder(150);

            // First byte is compressed 2 values.
            int first = buffer[0] / 40;   // first octet
            int second = buffer[0] % 40;  // second octet
            resultStr.Append(first + "." + second);
            offset++;
            headerLength--;

            while (headerLength > 0)
            {
                int result = 0;

                // this is where we decode individual values
                {
                    if ((buffer[offset] & HIGH_BIT) == 0)
                    {
                        // short encoding
                        result = (int)buffer[offset];
                        offset += 1;
                        --headerLength;
                    }
                    else
                    {
                        // long encoding
                        MutableByte tmp = new MutableByte();
                        bool completed = false;
                        do
                        {
                            tmp.Append((byte)(buffer[offset] & ~HIGH_BIT));
                            if ((buffer[offset] & HIGH_BIT) == 0)
                                completed = true;
                            offset += 1; // advance offset
                            --headerLength; // take out the processed byte from the header length
                        } while (!completed);

                        // convert byte array to integer
                        for (int i = 0; i < tmp.Length; i++)
                        {
                            result <<= 7;
                            result |= tmp[i];
                        }
                    }

                    resultStr.Append("." + result);
                }
            }

            return resultStr.ToString();   // decoded number
        }

        #endregion

        #region Hex dump logging.

        /// <summary>
        /// Log the given buffer, displaying printable characters
        /// </summary>
        public static void ByteStringLogit(byte[] buffer, LogLevel lvl, int indent, BYTE_LIST_TYPE blt)
        {
            int len = buffer.Length;

            if (buffer == null || buffer.Length < 1)
            {
                mLogger.Info("ByteStringLogit: Null or empty buffer, returning");
                return;
            }

            // Set num of spaces for indentation
            string spc = SpaceString(indent);

            int i;
            StringBuilder hexPart = new StringBuilder(48);
            StringBuilder charPart = new StringBuilder(10);

            // For OID and IPADDR, charPart needs to be built up front
            if (blt == BYTE_LIST_TYPE.OID)
            {
                charPart.Append(OidString(buffer));
            }
            else if (blt == BYTE_LIST_TYPE.IPADDR)
            {
                for (i = 0; i < len; i++)
                {
                    if (i != 0)
                        charPart.Append(".");
                    charPart.Append(buffer[i]);
                }
            }
            else if (blt == BYTE_LIST_TYPE.TIMETICKS)
            {
                charPart.Append(TimeTickString(buffer));
            }

            for (i = 0; i < len; i++)
            {
                if (i % 10 == 0 && i > 0)    // restart for new line
                {
                    mLogger.Info(spc + String.Format("{0,-48}    {1}", hexPart, charPart), lvl);
                    hexPart.Length = 0;
                    charPart.Length = 0;
                    if (blt == BYTE_LIST_TYPE.IPADDR || blt == BYTE_LIST_TYPE.OID)
                        charPart.Clear();   // only want on first line
                }

                hexPart.Append(String.Format("  {0:x2}", buffer[i]));

                if (!(blt == BYTE_LIST_TYPE.IPADDR || blt == BYTE_LIST_TYPE.OID || blt == BYTE_LIST_TYPE.TIMETICKS))
                {
                    if (buffer[i] >= 32 && buffer[i] < 127)
                        charPart.Append((char)buffer[i]);
                    else
                        charPart.Append(".");
                }
            }

            if (hexPart.Length > 0)   // add ending newline
                mLogger.Info(spc + String.Format("{0,-48}    {1}", hexPart, charPart), lvl);
        }

        #endregion

        #region SnmpComment and subroutines

        /// <summary>
        /// Load the comment fields with meanings appropriate for at least an Snmpv1 record
        /// </summary>
        public void SnmpComment()
        {
            int i;

            mLogger.Trace("SnmpComment: start");

            // It must start at a sequence
            if (_Type != ASN_SEQUENCE)
            {
                mLogger.Info("SnmpComment: not a sequence, returning");
                return;
            }

            for (i = 0; i < _ChildList.Count; i++)
            {
                SnmpTlv item = _ChildList[i];

                mLogger.Debug("SnmpComment:2: i=" + i + ", " + item.TypeString());

                if (i == 0)  // First item is the snmp version
                {
                    if (item._Type == ASN_INT)
                    {
                        switch (item._Packet[0])
                        {
                            case 0: item._Comment = "Version: SnmpV1"; break;
                            case 1: item._Comment = "Version: SnmpV2"; break;
                            case 2: item._Comment = "Version: SnmpV3"; break;
                            default: mLogger.Info("SnmpComment: not a known Snmp version");
                                break;
                        }
                    }
                }
                else if (i == 1)  // Second item is community string
                {
                    if (item._Type == ASN_OCTSTR)
                    {
                        item._Comment = "Community String";
                    }
                }
                else if (i > 1 && (item._Type == ASN_TRAPPDU || item._Type == ASN_V2TRAPPDU))  // Third and following items need to be PDU's
                {
                    item.TrapPduComment();
                }
            }
        }

        /// <summary>
        /// Fill out comments for TRAPPDU section
        /// </summary>
        public void TrapPduComment()
        {
            mLogger.Trace("TrapPduComment: start");

            // Must start at a sequence
            switch (_Type)
            {
                case ASN_TRAPPDU: V1TrapComment(); break;
                case ASN_V2TRAPPDU: V2TrapComment(); break;
                default: mLogger.Info("TrapPduComment: not a TrapPdu, returning"); break;
            }
        }

        /// <summary>
        /// Comment the PDU according to SNMP V1
        /// </summary>
        public void V1TrapComment()
        {
            int i;

            for (i = 0; i < _ChildList.Count; i++)
            {
                SnmpTlv item = _ChildList[i];

                mLogger.Trace("V1TrapComment:2: i=" + i + ", " + item.TypeString());

                if (i == 0)  // First item is the Enterprise OID
                {
                    if (item._Type == ASN_OBJID)
                    {
                        item._Comment = "Enterprise";
                    }
                }
                else if (i == 1)  // Second item is the From Device
                {
                    if (item._Type == ASN_IPADDR)
                    {
                        item._Comment = "From Device";
                    }
                }
                else if (i == 2)  // Third item is the Generic Trap type
                {
                    if (item._Type == ASN_INT)
                    {
                        item._Comment = "Generic Trap: " + GenTrapString(item._Packet[0]);
                    }
                }
                else if (i == 3)  // Fourth item is the Specific Trap type
                {
                    if (item._Type == ASN_INT)
                    {
                        if (item._Packet[0] == 0)
                            item._Comment = "Specific Trap - Unused";
                        else
                            item._Comment = "Specific Trap";
                    }
                }
                else if (i == 4)  // Fifth item is the time stamp
                {
                    if (item._Type == ASN_TIMETICKS)
                    {
                        item._Comment = "Timestamp";
                    }
                }
                else if (i == 5)  // Sixth item is the VarBindList sequence
                {
                    if (item._Type == ASN_SEQUENCE)
                    {
                        item.VarBindComment();
                    }
                }
            }
        }

        public void V2TrapComment()
        {
            int i;

            for (i = 0; i < _ChildList.Count; i++)
            {
                SnmpTlv item = _ChildList[i];

                mLogger.Trace("V2TrapComment:2: i=" + i + ", " + item.TypeString());

                if (i == 0)  // First item is the Request ID
                {
                    if (item._Type == ASN_INT)
                    {
                        item._Comment = "Request ID - Unused";
                    }
                }
                else if (i == 1)  // Second item is the error status - only used by response
                {
                    if (item._Type == ASN_INT)
                    {
                        item._Comment = "Error Status - Unused";
                    }
                }
                else if (i == 2)  // Third item is the error index - only used by response
                {
                    if (item._Type == ASN_INT)
                    {
                        item._Comment = "Error Index - Unused";
                    }
                }
                else if (item._Type == ASN_SEQUENCE)
                {
                    item.VarBindComment();
                }
            }
        }

        /// <summary>
        /// Fill out comments for VarBindList section
        /// </summary>
        private void VarBindComment()
        {
            int i;

            mLogger.Trace("VarBindComment: start");
            _Comment = "VarBind List";

            // All of it's sequences will be VarBind items
            for (i = 0; i < _ChildList.Count; i++)
            {
                SnmpTlv item = _ChildList[i];

                if (item._Type == ASN_SEQUENCE)
                {
                    item._Comment = "VarBind Item";
                }
            }
        }

        #endregion
        
        /// <summary>
        /// Confirm correct operation for this class.
        /// </summary>
        static public void Tester()
        {
            SnmpTlv tlv = null;
            mLogger.Info("SnmpTlv.Tester: start");

            // empty SNMP trap
            /*
            byte[] pkt1 = { 0x30, 0x23, 0x02, 0x01, 0x00,
                              0x04, 0x06, 0x70, 0x75, 0x62, 0x6c, 0x69, 0x63,
                              0xa4, 0x16, 0x06, 0x00,
                              0x40, 0x04, 0x0a, 0x00, 0x01, 0xaf,
                              0x02, 0x01, 0x00,
                              0x02, 0x01, 0x00,
                              0x43, 0x04, 0x02, 0x1d, 0x29, 0x20, 0x30, 0x00
                          };
            */

            /*
            // Full test trap
            byte[] pkt1 = { 0x30, 0x82, 0x01, 0x7b, // Sequence
                 0x02, 0x01, 0x00, // snmpv1
                 0x04, 0x06, 0x70,
                 0x75, 0x62, 0x6c, 0x69, 0x63, 0xa4, 0x82, 0x01, 0x6c, 0x06,
                 0x07, 0x2b, 0x06, 0x01, 0x04, 0x01, 0xb5, 0x69, 0x40, 0x04,
                 0x7f, 0x00, 0x00, 0x01, 0x02, 0x01, 0x03, 0x02, 0x01, 0x00,
                 0x43, 0x05, 0x00, 0xe2, 0x56, 0x7d, 0xc9, 0x30, 0x82, 0x01,
                 0x4c, 0x30, 0x1a, 0x06, 0x0e, 0x2b, 0x06, 0x01, 0x04, 0x01,
                 0xb5, 0x69, 0x02, 0x08, 0x02, 0x01, 0x01, 0x01, 0x04, 0x04,
                 0x08, 0x43, 0x49, 0x4d, 0x53, 0x20, 0x49, 0x6e, 0x63, 0x30,
                 0x1c, 0x06, 0x0e, 0x2b, 0x06, 0x01, 0x04, 0x01, 0xb5, 0x69,
                 0x02, 0x08, 0x01, 0x04, 0x06, 0x01, 0x12, 0x04, 0x0a, 0x38,
                 0x39, 0x39, 0x39, 0x30, 0x30, 0x30, 0x30, 0x30, 0x31, 0x30,
                 0x34, 0x06, 0x0e, 0x2b, 0x06, 0x01, 0x04, 0x01, 0xb5, 0x69,
                 0x02, 0x08, 0x01, 0x04, 0x06, 0x01, 0x11, 0x04, 0x22, 0x46,
                 0x50, 0x41, 0x3a, 0x30, 0x30, 0x30, 0x30, 0x30, 0x3a, 0x30,
                 0x35, 0x32, 0x36, 0x31, 0x33, 0x35, 0x32, 0x30, 0x34, 0x3a,
                 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,
                 0x3a, 0x3a, 0x4e, 0x30, 0x13, 0x06, 0x0e, 0x2b, 0x06, 0x01,
                 0x04, 0x01, 0xb5, 0x69, 0x02, 0x08, 0x01, 0x04, 0x06, 0x01,
                 0x01, 0x04, 0x01, 0x41, 0x30, 0x15, 0x06, 0x0e, 0x2b, 0x06,
                 0x01, 0x04, 0x01, 0xb5, 0x69, 0x02, 0x08, 0x01, 0x04, 0x06,
                 0x01, 0x03, 0x04, 0x03, 0x55, 0x50, 0x47, 0x30, 0x13, 0x06,
                 0x0e, 0x2b, 0x06, 0x01, 0x04, 0x01, 0xb5, 0x69, 0x02, 0x08,
                 0x01, 0x04, 0x06, 0x01, 0x04, 0x04, 0x01, 0x31, 0x30, 0x15,
                 0x06, 0x0e, 0x2b, 0x06, 0x01, 0x04, 0x01, 0xb5, 0x69, 0x02,
                 0x08, 0x01, 0x04, 0x06, 0x01, 0x06, 0x04, 0x03, 0x4d, 0x41,
                 0x4a, 0x30, 0x19, 0x06, 0x0e, 0x2b, 0x06, 0x01, 0x04, 0x01,
                 0xb5, 0x69, 0x02, 0x08, 0x01, 0x04, 0x06, 0x01, 0x1a, 0x04,
                 0x07, 0x30, 0x2e, 0x30, 0x2e, 0x30, 0x2e, 0x30, 0x30, 0x29,
                 0x06, 0x0e, 0x2b, 0x06, 0x01, 0x04, 0x01, 0xb5, 0x69, 0x02,
                 0x08, 0x01, 0x04, 0x06, 0x01, 0x1b, 0x04, 0x17, 0x41, 0x6c,
                 0x61, 0x72, 0x6d, 0x54, 0x72, 0x61, 0x71, 0x20, 0x50, 0x72,
                 0x6f, 0x78, 0x79, 0x20, 0x4d, 0x65, 0x73, 0x73, 0x61, 0x67,
                 0x65, 0x30, 0x12, 0x06, 0x0e, 0x2b, 0x06, 0x01, 0x04, 0x01,
                 0xb5, 0x69, 0x02, 0x08, 0x01, 0x04, 0x06, 0x01, 0x1c, 0x04,
                 0x00, 0x30, 0x28, 0x06, 0x0e, 0x2b, 0x06, 0x01, 0x04, 0x01,
                 0xb5, 0x69, 0x02, 0x08, 0x01, 0x04, 0x06, 0x01, 0x63, 0x04,
                 0x16, 0x43, 0x49, 0x4d, 0x53, 0x20, 0x54, 0x65, 0x73, 0x74,
                 0x20, 0x41, 0x6c, 0x61, 0x72, 0x6d, 0x20, 0x74, 0x79, 0x70,
                 0x65, 0x20, 0x31  };
            */

            // Caused an exception
            byte[] pkt1 = {
                0x30,  0x22,  0x02,  0x01,  0x00,  0x04,  0x06,  0x70,  0x75,  0x62,
                0x6c,  0x69,  0x63,  0xa4,  0x15,  0x06,  0x00,  0x40,  0x04,  0x0a,
                0x00,  0x01,  0xaf,  0x02,  0x01,  0x00,  0x02,  0x01,  0x00,  0x43,
                0x03,  0x22,  0x4b,  0x52,  0x30,  0x00
                };

            // Snmp v2 packet
            byte[] pkt2 = {
                0x30, 0x82, 0x02, 0x3b, 0x02, 0x01, 0x01, 0x04, 0x06, 0x70,
                0x75, 0x62, 0x6c, 0x69, 0x63, 0xa7, 0x82, 0x02, 0x2c, 0x02,
                0x01, 0x00, 0x02, 0x01, 0x00, 0x02, 0x01, 0x00, 0x30, 0x82,
                0x02, 0x1f, 0x30, 0x10, 0x06, 0x08, 0x2b, 0x06, 0x01, 0x02,
                0x01, 0x01, 0x03, 0x00, 0x43, 0x04, 0x01, 0x77, 0x08, 0x4e,
                0x30, 0x1a, 0x06, 0x0a, 0x2b, 0x06, 0x01, 0x06, 0x03, 0x01,
                0x01, 0x04, 0x01, 0x00, 0x06, 0x0c, 0x2b, 0x06, 0x01, 0x04,
                0x01, 0xb5, 0x69, 0x02, 0x46, 0x01, 0x02, 0x01, 0x30, 0x3c,
                0x06, 0x0d, 0x2b, 0x06, 0x01, 0x04, 0x01, 0xb5, 0x69, 0x02,
                0x46, 0x01, 0x01, 0x01, 0x00, 0x04, 0x2b, 0x32, 0x30, 0x30,
                0x30, 0x30, 0x34, 0x39, 0x30, 0x35, 0x30, 0x20, 0x33, 0x31,
                0x2f, 0x31, 0x31, 0x3a, 0x33, 0x38, 0x2c, 0x43, 0x4c, 0x52,
                0x7c, 0x41, 0x4c, 0x4c, 0x20, 0x41, 0x4c, 0x41, 0x52, 0x4d,
                0x53, 0x20, 0x52, 0x45, 0x53, 0x4f, 0x4c, 0x56, 0x45, 0x44,
                0x30, 0x1a, 0x06, 0x0d, 0x2b, 0x06, 0x01, 0x04, 0x01, 0xb5,
                0x69, 0x02, 0x46, 0x01, 0x01, 0x03, 0x00, 0x04, 0x09, 0x32,
                0x37, 0x38, 0x37, 0x30, 0x36, 0x34, 0x34, 0x38, 0x30, 0x18,
                0x06, 0x0a, 0x2b, 0x06, 0x01, 0x06, 0x03, 0x01, 0x01, 0x04,
                0x03, 0x00, 0x06, 0x0a, 0x2b, 0x06, 0x01, 0x04, 0x01, 0xb5,
                0x69, 0x02, 0x46, 0x01, 0x30, 0x19, 0x06, 0x08, 0x2b, 0x06,
                0x01, 0x02, 0x01, 0x01, 0x06, 0x00, 0x04, 0x0d, 0x53, 0x52,
                0x41, 0x20, 0x4d, 0x4c, 0x4d, 0x20, 0x41, 0x4c, 0x41, 0x52,
                0x4d, 0x30, 0x15, 0x06, 0x0d, 0x2b, 0x06, 0x01, 0x04, 0x01,
                0xb5, 0x69, 0x02, 0x46, 0x02, 0x01, 0x01, 0x00, 0x40, 0x04,
                0x0a, 0x64, 0x11, 0x0d, 0x30, 0x3c, 0x06, 0x0d, 0x2b, 0x06,
                0x01, 0x04, 0x01, 0x8f, 0xb6, 0x60, 0x02, 0x46, 0x01, 0x01,
                0x01, 0x04, 0x2b, 0x54, 0x68, 0x65, 0x20, 0x48, 0x61, 0x72,
                0x74, 0x66, 0x6f, 0x72, 0x64, 0x20, 0x40, 0x20, 0x42, 0x6f,
                0x75, 0x6c, 0x64, 0x65, 0x72, 0x2c, 0x20, 0x43, 0x4f, 0x20,
                0x2d, 0x20, 0x43, 0x4d, 0x32, 0x20, 0x2d, 0x20, 0x4d, 0x41,
                0x53, 0x20, 0x53, 0x75, 0x70, 0x31, 0x30, 0x1b, 0x06, 0x0d,
                0x2b, 0x06, 0x01, 0x04, 0x01, 0x8f, 0xb6, 0x60, 0x02, 0x46,
                0x01, 0x01, 0x02, 0x04, 0x0a, 0x32, 0x30, 0x30, 0x30, 0x30,
                0x34, 0x39, 0x30, 0x35, 0x30, 0x30, 0x17, 0x06, 0x0d, 0x2b,
                0x06, 0x01, 0x04, 0x01, 0x8f, 0xb6, 0x60, 0x02, 0x46, 0x01,
                0x01, 0x04, 0x04, 0x06, 0x31, 0x33, 0x37, 0x31, 0x35, 0x37,
                0x30, 0x16, 0x06, 0x0d, 0x2b, 0x06, 0x01, 0x04, 0x01, 0x8f,
                0xb6, 0x60, 0x02, 0x46, 0x01, 0x01, 0x05, 0x04, 0x05, 0x41,
                0x55, 0x44, 0x49, 0x58, 0x30, 0x59, 0x06, 0x0d, 0x2b, 0x06,
                0x01, 0x04, 0x01, 0x8f, 0xb6, 0x60, 0x02, 0x46, 0x01, 0x01,
                0x06, 0x04, 0x48, 0x53, 0x65, 0x65, 0x20, 0x44, 0x6f, 0x63,
                0x75, 0x6d, 0x65, 0x6e, 0x74, 0x73, 0x20, 0x69, 0x63, 0x6f,
                0x6e, 0x20, 0x66, 0x6f, 0x72, 0x20, 0x63, 0x6f, 0x6e, 0x74,
                0x61, 0x63, 0x74, 0x20, 0x69, 0x6e, 0x66, 0x6f, 0x72, 0x6d,
                0x61, 0x74, 0x69, 0x6f, 0x6e, 0x2e, 0x0d, 0x0a, 0x0d, 0x0a,
                0x55, 0x73, 0x65, 0x20, 0x53, 0x41, 0x4c, 0x20, 0x66, 0x6f,
                0x72, 0x20, 0x63, 0x6f, 0x6e, 0x6e, 0x65, 0x63, 0x74, 0x69,
                0x76, 0x69, 0x74, 0x79, 0x2e, 0x30, 0x62, 0x06, 0x0d, 0x2b,
                0x06, 0x01, 0x04, 0x01, 0x8f, 0xb6, 0x60, 0x02, 0x46, 0x01,
                0x01, 0x07, 0x04, 0x51, 0x41, 0x6c, 0x6c, 0x20, 0x41, 0x6c,
                0x61, 0x72, 0x6d, 0x73, 0x20, 0x52, 0x65, 0x73, 0x6f, 0x6c,
                0x76, 0x65, 0x64, 0x20, 0x69, 0x73, 0x20, 0x73, 0x65, 0x6e,
                0x74, 0x20, 0x77, 0x68, 0x65, 0x6e, 0x20, 0x6e, 0x6f, 0x20,
                0x6d, 0x6f, 0x72, 0x65, 0x20, 0x61, 0x63, 0x74, 0x69, 0x76,
                0x65, 0x20, 0x61, 0x6c, 0x61, 0x72, 0x6d, 0x73, 0x20, 0x61,
                0x72, 0x65, 0x20, 0x70, 0x72, 0x65, 0x73, 0x65, 0x6e, 0x74,
                0x20, 0x6f, 0x6e, 0x20, 0x74, 0x68, 0x65, 0x20, 0x73, 0x79,
                0x73, 0x74, 0x65, 0x6d, 0x2e
                          };
            // Create the TLV of packet
            try
            {
                tlv = new SnmpTlv(pkt2, false);
            }
            catch (Exception ex)
            {
                mLogger.Info("SnmpTlv.Tester: constructor returned ex=" + ex);
                return;
            }

            try
            {
                tlv.SnmpComment();
                mLogger.Info("");   // Blank line to separate
                tlv.Logit(LogLevel.Info, 0);
            }
            catch (Exception ex)
            {
                mLogger.Error("SnmpTlv.Tester: Logit returned ex=" + ex);
                return;
            }

            mLogger.Info("SnmpTlv.Tester: returning");
        }

        /// <summary>
        /// Return string equivalent of ASN.1 value
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public string AsnToString(byte value)
        {
            switch (value)
            {
                case ASN_INT: return "INTEGER";
                case ASN_OCTSTR: return "OCTSTR";
                case ASN_NULL: return "NULL";
                case ASN_OBJID: return "OID";
                case ASN_SEQUENCE: return "SEQ";
                case ASN_IPADDR: return "IPADDR";
                case ASN_COUNTER: return "COUNTER";
                case ASN_GAUGE: return "GAUGE";
                case ASN_TIMETICKS: return "TIMETICKS";
                case ASN_OPAQUE: return "OPAQUE";
                case ASN_NSAP_ADDRESS: return "NSAP ADDRESS";
                case ASN_COUNTER64: return "COUNTER64";
                case ASN_UINTEGER32: return "UINTEGER32";
                case ASN_GETREQPDU: return "GETREQ PDU";
                case ASN_GETRESPPDU: return "GETRESP PDU";
                case ASN_TRAPPDU: return "TRAP PDU";
                case ASN_V2GETBULKPDU: return "V2 GetBulk Request PDU";
                case ASN_V2INFORMPDU: return "V2 Inform Request PDU";
                case ASN_V2TRAPPDU: return "V2 TRAP PDU";
                default: return "Unknown "+value;
            }
        }
    }
}
