﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Sockets;
using System.Linq;
using System.Text;
using NLog;
using System.Threading;

namespace AlarmReader
{
    class SocketSupport
    {
        #region Constants

        public static string CLASS_NAME = typeof(SocketSupport).ToString();

        public delegate void ReceiveDelegateH(IAsyncResult result);
        public delegate void ReceiveDelegateE();

        public const string MS_DATE_FORMAT = "yyyy-MM-dd h:mm:ss.fff tt";
        public const int RESTART_MAX = 10;         // Maximum num of restarts allowed
        public const int INP_BUF_SIZE = 64 * 1024;    // 64k socket read buffer

        #endregion

        #region Private Members

        // Set to true for testing error during run (fail after first alarm), false for normal
        bool _intentionalFail = false;
        int _intentionalFailCount = 0;

        ReceiveDelegateH _ReceiveCallback;           // Tell caller we received a message
        ReceiveDelegateE _ErrorCallback;            // Tell caller we need to shut down

        protected byte[] _inbuffer;
        protected int _inbufferLength;

        protected Socket _ReceiveSocket;   // Sock to be binded to
        protected IPEndPoint _peerIP;           // Really needs to be class variable
        private int _ReceivePort;                  // Port to listen on

        protected bool _isShuttingDown;           // True if we are in process of shutdown, false otherwise
        protected bool _isRunning;

        private int _restartCount;

        private static Logger mLogger = LogManager.GetCurrentClassLogger();

        #endregion Private Members

        #region Properties

        public byte[] InBuffer
        {
            get { return _inbuffer; }
            set { _inbuffer = value; }
        }

        public int InBufferLength
        {
            get { return _inbufferLength; }
            set { _inbufferLength = value; }
        }

        public Socket ReceiveSocket
        {
            get { return _ReceiveSocket; }
            set { _ReceiveSocket = value; }
        }

        public int ReceivePort
        {
            get { return _ReceivePort; }
            set { _ReceivePort = value; }
        }

        public bool IsShuttingDown
        {
            get { return _isShuttingDown; }
            set { _isShuttingDown = value; }
        }

        public bool IsRunning
        {
            get { return _isRunning; }
            set { _isRunning = value; }
        }

        #endregion

        public SocketSupport()
        {
            try
            {
                // Initial values
                _inbuffer = null;
                _inbufferLength = -1;
                _restartCount = 0;
                _ReceiveSocket = null;
                _ReceivePort = -1;

                _isShuttingDown = false;
                _isRunning = false;


            }
            catch (Exception ex)  // If all else fails
            {
                Exception newEx = new Exception("SocketSupport constructor failed", ex);
                throw newEx;
            }
        }

        /// <summary>
        /// Sets the callback for receiving UDP packets.
        /// </summary>
        /// <param name="callerDelegate">Informs caller we have stopped</param>
        public void SetCallbackH(ReceiveDelegateH callerDelegate, ReceiveDelegateE errorDelegate)
        {
            _ReceiveCallback = callerDelegate;
            _ErrorCallback = errorDelegate;
        }


        #region Socket manipulation

        /// <summary>
        /// Hit a socket problem. Restart the service
        /// </summary>
        /// <returns>True if restart succeeded, false if failed</returns>
        public bool RestartSocket()
        {
            // Don't do anything during shutdown
            if (_isShuttingDown == true)
            {
                mLogger.Info("Cancelling RestartSocket() during shutdown");
                return true;
            }

            // Shut down existing socket, if needed
            if (_ReceiveSocket != null)
            {
                _ReceiveSocket.Close();
                _ReceiveSocket = null;
            }

            // Update the restart count. Want to shut down on certain number of restarts
            _restartCount++;
            mLogger.Info("RestartSocket: start, count is now " + _restartCount + ", time=" +
                DateTime.Now.ToString(MS_DATE_FORMAT));

            if (_restartCount > RESTART_MAX)
            {
                return false;
            }

            // Open the socket, actual process done by callback function
            bool result = OpenBindSocket();

            if (result == false)
            {
                mLogger.Error("RestartSocket: failed to open or bind socket!");
                Thread.Sleep(100);  // 100ms
                return false;
            }

            mLogger.Info("restartSocket: successfully bound socket");

            // Register to receive packets. This must be done each time we get one.
            if (!RegisterReceiveOperation())
            {
                mLogger.Error("RestartSocket: RegisterReceiveOperation() failed!");

                return false;
            }

            mLogger.Info("restartSocket: success RegisterReceiveOperation");
            _restartCount = 0;   // success resets the count
            return true;
        }

        /// <summary>
        /// Open the socket, bind it to the receive port.
        /// </summary>
        /// <returns>True on success, false on failure</returns>
        private bool OpenBindSocket()
        {
            // Create IP/UDP socket
            try
            {
                _ReceiveSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            }
            catch (Exception ex)
            {
                mLogger.Error("OpenBindSocket: socket initialization failed with error: " + ex.Message);

                // there is no need to close the socket because it was never correctly created
                _ReceiveSocket = null;
                return false;
            }

            mLogger.Info("Binding socket to local port " + _ReceivePort);

            // prepare to "bind" the socket to the local port number
            // binding notifies the operating system that application 
            // wishes to receive data sent to the specified port number
            try
            {
                // prepare EndPoint that will bind the application to all available IP addresses
                EndPoint localEP = new IPEndPoint(IPAddress.Any, _ReceivePort);

                // bind socket
                _ReceiveSocket.Bind(localEP);
            }
            catch (Exception ex)
            {
                mLogger.Error("OpenBindSocket: Receiver Bind failed with error: " + ex.Message);

                _ReceiveSocket.Close();
                _ReceiveSocket = null;
                return false;
            }

            // Return success
            return true;
        }

        /// <summary>
        /// Set up to receive a UDP packet
        /// </summary>
        /// <returns>Returns true on success, false on fail</returns>
        public bool RegisterReceiveOperation()
        {
            mLogger.Trace("RegisterReceiveOperation: ReceiveSocket=" + _ReceiveSocket);

            if (_ReceiveSocket == null)
            {
                mLogger.Error("RegisterReceiveOperation: attempted to use closed socket!");
                return false; // socket has been closed
            }
            try
            {
                _peerIP = new IPEndPoint(IPAddress.Any, 0); // receive from anybody
                EndPoint ep = (EndPoint)_peerIP;

                // Possible memory leak?
                _inbuffer = new byte[INP_BUF_SIZE]; // nice and big receive buffer

                if (_intentionalFail == true && (++_intentionalFailCount > 1))
                    throw new Exception("Intentional fail exception");   // test

                mLogger.Debug("RegisterReceiveOperation:2: About to call BeginReceiveFrom");

                _ReceiveSocket.BeginReceiveFrom(_inbuffer, 0, INP_BUF_SIZE, SocketFlags.None, ref ep,
                    new AsyncCallback(ReceiveCallback), _ReceiveSocket);

                mLogger.Debug("RegisterReceiveOperation:3: Called BeginReceiveFrom success");

            }
            catch (Exception ex)
            {
                mLogger.Error("RegisterReceiveOperation: failed with message: " + ex.Message);

                _ReceiveSocket.Close();
                _ReceiveSocket = null;
                Thread.Sleep(100);  // 100ms
                return false;
            }

            // Success
            mLogger.Info("RegisterReceiveOp: Successfully registered ReceiveCallback");
            return true;
        }

        /// <summary>
        /// This is called each time a UDP packet is received on our receive port.
        /// We process it and set up for the next receive.
        /// </summary>
        /// <param name="result">Contains socket and other info.</param>
        private void ReceiveCallback(IAsyncResult result)
        {
            // If service is shutting down, we get odd calls here as socket is closed. Ignore
            if (_isShuttingDown == true)
            {
                mLogger.Info("ReceiveCallback returning immediately because in shutdown process");
                return;
            }

            // Note Entry
            mLogger.Info("\r\n"); // blank before new event
            mLogger.Info("ReceiveCallback: ======= start =======  Packet received");

            // Caller's handler
            _ReceiveCallback(result);

            // Always must be last, can cause another ReceiveCallback!
            if (RegisterReceiveOperation() == false)
            {
                if (RestartSocket() == false)
                {
                    _ErrorCallback();   // Shut down calling manager
                }
            }

            mLogger.Info("======= End of ReceiveCallback =======");
        }

        /// <summary>
        /// This completes the read of the UDP packet and checks for error.
        /// </summary>
        /// <param name="result">Information to complete the read</param>
        /// <returns>Number of bytes in packet, or -1 on error</returns>
        public int CompleteUdpRead(IAsyncResult result)
        {
            // Return if our socket is closed, somehow snuck in
            if (_ReceiveSocket == null)
            {
                mLogger.Error("CompleteUdpRead: received over null socket, discarding");
                _inbufferLength = -1;   // failed
                return -1;
            }

            // Init
            IPEndPoint _peerIP = new IPEndPoint(IPAddress.Any, 0);

            // Finish up the UDP receive
            try
            {
                EndPoint ep = (EndPoint)_peerIP;

                // Internally, it sets _inbuffer to received characters
                _inbufferLength = _ReceiveSocket.EndReceiveFrom(result, ref ep);
                _peerIP = (IPEndPoint)ep;
            }
            catch (Exception ex)
            {
                mLogger.Error("CompleteUdpRead: Receive operation failed with message: " + ex.Message);
                _inbufferLength = -1;   // failed
            }

            return _inbufferLength;
        }

        #endregion

        public void Stop()
        {
            _isShuttingDown = true;   // Stop any socket restarts

            mLogger.Info(CLASS_NAME+".Stop: Support stopped at " +
                DateTime.Now.ToString(MS_DATE_FORMAT));

            try
            {
                if (_ReceiveSocket != null)
                {
                    _ReceiveSocket.Close();
                    _ReceiveSocket = null;
                }
            }
            catch { }

            _isRunning = false;

        }
    }
}
