﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;

using SNMPSharp;
using NLog;

namespace AlarmReader
{
    public static class TestAlarm
    {
        public const string MS_DATE_FORMAT = "yyyy-MM-dd HH:mm:ss.fff";
        public const string RETURN_OID = "1.3.6.1.4.1.1111";
        public const int HIGHEST_SNMP_ALARM_TYPE = 7;    // needs to agree with case statement below.

        static BackgroundWorker backgroundWorker = null;
        static int averageInterval;
        static string serverName;
        static int serverPort;

        private static Logger mLogger = LogManager.GetCurrentClassLogger();

        #region Send a Series of test alarms via SNMP

        /// <summary>
        /// Send a random series of Test alarms at random intervals averaging given milliseconds to desired server
        /// </summary>
        /// <param name="avgInterval">milliseconds between alarms</param>
        /// <param name="svName">Server to send do</param>
        public static void SendSeries(int avgInterval, string svName, int svPort)
        {
            // See if there is already a series running
            if (backgroundWorker != null)
            {
                mLogger.Info("SendSeries: already an alarm series running, returning");
                return;
            }

            // Set our variables
            averageInterval = avgInterval;
            serverName = svName;
            serverPort = svPort;

            // Generate our background thread
            try
            {
                backgroundWorker = new BackgroundWorker();
                backgroundWorker.DoWork += new DoWorkEventHandler(backgroundWorker_DoWork);
                backgroundWorker.WorkerSupportsCancellation = true;
                backgroundWorker.RunWorkerAsync();
                mLogger.Info("TestAlarm.Series Started.");
            }
            catch (Exception ex)
            {
                mLogger.Error("SendSeries: Exception on startup="+ex);
            }
        }

        /// <summary>
        /// Stop any currently running series
        /// </summary>
        public static void StopSeries()
        {
            // See if there is already a series running
            if (backgroundWorker == null)
            {
                mLogger.Info("StopSeries: no alarm series running, returning");
                return;
            }

            // Stop it
            try
            {
                backgroundWorker.CancelAsync();
                System.Threading.Thread.Sleep(510);	// wait for background thread to stop
                mLogger.Info("TestAlarm.Series Completed.");
                backgroundWorker = null;
            }
            catch (Exception ex)
            {
                mLogger.Error("StopSeries: Exception on stop=" + ex);
            }
        }

        /// <summary>
        /// Do the alarms in this background thread.
        /// </summary>
        static void backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            int alarmType, sleepVal, dur;
            BackgroundWorker bw = (BackgroundWorker)sender;

            // Loop through generating alarms

            while (!bw.CancellationPending)
            {
                // Pick a random alarm
                alarmType = MainWindow.RandomInt(1, HIGHEST_SNMP_ALARM_TYPE);

                // Send it
                dur = Send1SnmpAlarm(alarmType, serverName, MainWindow.ReceivePort);

                if (dur < 0)
                {
                    mLogger.Info("SendSeries: failed to send test snmp alarm at " + DateTime.Now.ToString(MS_DATE_FORMAT));
                    return;
                }

                // Wait a random time to continue
                if (!bw.CancellationPending)
                {
                    sleepVal = MainWindow.RandomInt(1, 2 * averageInterval) - dur;
                    if (sleepVal < 1) sleepVal = 1;

                    System.Threading.Thread.Sleep(sleepVal);
                }
            }
        }

        #endregion

        #region Send 1 SNMP Alarm

        /// <summary>
        /// This sends 1 individual alarm of the specified type to the specified server.
        /// </summary>
        /// <param name="alarmType">Internal alarm type 1-x</param>
        /// <param name="serverName">Server to send to</param>
        /// <returns>num of ms on success, -1 on failure</returns>
        public static int Send1SnmpAlarm(int alarmType, string serverName, int aToPort)
        {
            TrapAgent agent = new TrapAgent();
            VbCollection col;
            DateTime before, after;
            TimeSpan diff;
            IpAddress ipAddr = new IpAddress(serverName);

            // Create alarm type values
            mLogger.Info("Send1SnmpAlarm:1: alarmType=" + alarmType+", serverName='"+serverName+"'");
            before = DateTime.Now;
            col = CreateVbColl(alarmType);

            // Send trap to desired recipient
            bool rc = SendV1Trap(agent, ipAddr, aToPort, col, alarmType);

            after = DateTime.Now;
            diff = after - before;

            string result = "SUCCESS";
            if (rc == false) result = "FAIL";

            mLogger.Info("Send1SnmpAlarm: Sent alarm " + alarmType + " at " + before.ToString(MS_DATE_FORMAT) +
                    " to "+ serverName+" ("+ipAddr+") result (" + result + ")" +
                    ",  dur " + diff.TotalMilliseconds + " ms");

            // Act on result

            if (rc == false)
                return -1;
            else
                return (int) diff.TotalMilliseconds;
        }

        // Send the trap with our oid
        /// <summary>
        /// This sends a SNMPv1 trap to the requested server with our return address.
        /// </summary>
        /// <param name="agent">Trap agent to send with</param>
        /// <param name="toAddr">IP address (name) of desired server</param>
        /// <param name="upTime">Sender up time (distinguish messages)</param>
        /// <param name="col">Varbinditem collection making up the PDU</param>
        /// <param name="alarmType">Which alarm type we are sending (internal)</param>
        /// <returns>true on success, false on fail</returns>
        private static bool SendV1Trap(TrapAgent agent, IpAddress toAddr, int aToPort, VbCollection col, int alarmType)
        {
            // uT (sysUpTime) is number of hundreds of second since network reboot. We don't have network reboot time, so we just
            // assume 1/1/2010. Fortunately, the value doesn't seem to be really used.
            TimeSpan tS = DateTime.Now - new DateTime(2010, 1, 1);
            uint uT = (uint) (tS.TotalMilliseconds*10);

            try
            {
                // Note: Linkup is the generic trap (or Enterprise). The 0 after it is the specific trap. Since we're not a manufacturer,
                // we leave it zero. Fortunately, none of these values are really used by us.
                agent.SendV1Trap(toAddr, aToPort, "public",
                             new Oid(RETURN_OID), new IpAddress("localhost"),
                             SnmpConstants.LinkUp, 0, uT, col);
            }
            catch (Exception ex)
            {
                mLogger.Info("SendV1Trap: Failed to send trap for alarm type " + alarmType + " at " + 
                    DateTime.Now.ToString(MS_DATE_FORMAT) + ", ex='" + ex + "'");
                return false;
            }
            return true;
        }

        /// <summary>
        /// Create a new VbCollection, and load alarmtype data into it.
        /// </summary>
        /// <param name="alarmType"></param>
        /// <returns></returns>
        private static VbCollection CreateVbColl(int alarmType)
        {
            // Variable Binding collection to send with the trap
            VbCollection col = new VbCollection();

            switch (alarmType)
            {
                case 1:
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.2.1.1.1.4"), new OctetString("Test Inc"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.18"), new OctetString("8999000001"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.17"), new OctetString("FPA:00000:0526135204:0000000000::N"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.1"), new OctetString("A"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.3"), new OctetString("UPG"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.4"), new OctetString("1"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.6"), new OctetString("MAJ"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.26"), new OctetString("0.0.0.0"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.27"), new OctetString("Proxy Message"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.28"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.99"), new OctetString("Test Alarm type 1"));
                    break;

                case 2:
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.2.1.1.1.4"), new OctetString("Test Inc"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.18"), new OctetString("8999000002"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.17"), new OctetString("FPA:00000:0526140245:0000000000::N"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.1"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.3"), new OctetString("ALL ALARMS RESOLVED"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.4"), new OctetString("n"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.6"), new OctetString("WAR"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.26"), new OctetString("0.0.0.0"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.27"), new OctetString("Proxy Message"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.28"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.99"), new OctetString("Test Alarm type 2 - All Alarms Resolved"));
                    break;

                case 3:
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.2.1.1.1.4"), new OctetString("Test Inc"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.18"), new OctetString("8999000003"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.17"), new OctetString("FPA:00000:0526140347:0000000000::N"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.1"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.3"), new OctetString("VP-VOICE_PORT"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.4"), new OctetString("1"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.6"), new OctetString("MAJ"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.26"), new OctetString("0.0.0.0"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.27"), new OctetString("Proxy Message"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.28"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.99"), new OctetString("Test Alarm type 3 - VP-VOICE_PORT"));
                    break;

                case 4:
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.2.1.1.1.4"), new OctetString("Test Inc"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.18"), new OctetString("8999000004"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.17"), new OctetString("FPA:00000:0526140444:0000000000::N"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.1"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.3"), new OctetString("REBOOT"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.4"), new OctetString("n"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.6"), new OctetString("WAR"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.26"), new OctetString("0.0.0.0"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.27"), new OctetString("Proxy Message"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.28"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.99"), new OctetString("Test Alarm type 4 - REBOOT"));
                    break;

                case 5:
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.2.1.1.1.4"), new OctetString("Test Inc"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.18"), new OctetString("8999000005"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.17"), new OctetString("FPA:00000:0526140723:0000000000::N"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.1"), new OctetString("42"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.3"), new OctetString("VM-ALARM_ORIG"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.4"), new OctetString("0"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.6"), new OctetString("MIN"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.26"), new OctetString("0.0.0.0"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.27"), new OctetString("Proxy Message"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.28"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.99"), new OctetString("Test Alarm type 5 - Test of alarm origination"));
                    break;

                case 6:
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.2.1.1.1.4"), new OctetString("Test Inc"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.18"), new OctetString("8999000006"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.17"), new OctetString("FPA:00000:0526143023:0000000000::N"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.1"), new OctetString("11A02A"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.3"), new OctetString("DS1-FAC"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.4"), new OctetString("n"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.6"), new OctetString("MIN"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.26"), new OctetString("0.0.0.0"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.27"), new OctetString("Proxy Message"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.28"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.99"), new OctetString("Test Alarm type 6 - Long desc. DS1 Facility.\r\n\r\n" +
                        "The DS1-FAC is a part of the DS1 Converter (DS1C-BD) complex. The DS1C-BD complex consists of two TN574 DS1C-BD circuit packs or two TN1654 DS1C-BD circuit packs connected by one to four DS1 facilities. The MO name for the DS1 converter circuit pack is DS1C-BD, and the MO name for the connected DS1 facilities is DS1-FAC.\r\n\r\n" +
                        "The TN1654 DS1 Converter circuit pack is a redesign of the TN574 DS1 Converter board. The TN1654 provides functionality equivalent to the TN574, and adds a 32-channel E1 interface for international support as well as circuit-switched wideband connections (NxDS0). The TN1654 DS1 converter circuit pack is not compatible with the TN574 DS1 converter board or the Y-cable used to interface to the TN574 DS1 converter. A TN573B (or higher-suffixed) SNI board must be used when connecting to the TN1654 DS1 converter board. See DS1C-BD for information about both DS1 converter circuit packs.\r\n\r\n" +
                        "The DS1 converter complex is a part of the port-network connectivity (PNC). The DS1 converter complex is used to extend the range of the 32-Mbps fiber links that connect PNs to the Center Stage Switch, allowing PNs to be located at remote sites.\r\n\r\n" +
                        "The DS1 converter circuit pack contains on-board firmware that detects DS1 facility alarms and errors, communicates status to maintenance software, and runs either background or on-demand tests from maintenance software. The overall maintenance software strategy includes demand tests, recovery strategies, error and alarm logging, and periodic audits."));
                    break;

                case 7:
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.2.1.1.1.4"), new OctetString("Test Inc"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.18"), new OctetString("8999000007"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.17"), new OctetString("FPA:00000:0526150010:0000000000::N"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.1"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.3"), new OctetString("TESTING INADS LINK"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.4"), new OctetString("n"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.6"), new OctetString("WAR"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.26"), new OctetString("0.0.0.0"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.27"), new OctetString("Proxy Message"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.28"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.99"), new OctetString("Test Alarm type 7 - TEST LINK. MEDIA GATEWAY.\r\n\r\n" +
                        "The MEDIA GATEWAY maintenance object monitors the H.248 link to the Media Gateway. It logs errors when there are H.248 link problems or when hyperactive H.248 link bounce occurs."));
                    break;

                case 8:
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.2.1.2.1.9"), new OctetString("4"));  // Major severity
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.2.1.2.1.2"), new OctetString("07 DA 06 11 0A 35 1F 00")); // date/time
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.2.1.2.1.10"), new OctetString("8999000008"));  // DeviceID
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.2.1.2.1.5"), new OctetString("2"));  // AppEntity
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.8.1.4.6.1.99"), new OctetString("Test Alarm type 8 - IP Office"));
                    col.Add(new Oid("1.3.6.1.2.1.1.1"), new OctetString("IP 406 DS 5.0(15)"));  // Device desc
                    break;
                    
                case 9:   // Definity alarm without severity
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.70.1.1.1.0"), new OctetString("1000270571 08/01:19,EOF,CLR|ALL ALARMS RESOLVED"));  
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.70.1.1.3.0"), new OctetString("1307510359"));
                    col.Add(new Oid("1.3.6.1.2.1.1.6.0"), new OctetString("SRA MLM ALARM"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.70.2.1.1.0"), new OctetString("10.218.112.44"));
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.1"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.2"), new OctetString("1000270571"));
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.4"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.5"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.6"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.7"), new OctetString("All Alarms Resolved is sent when no more active alarms are present on the system."));
                    break;
                    
                case 10:   // Definity alarm with severity
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.70.1.1.1.0"), new OctetString("1000270571 08/01:20,EOF,ACT|,SP-REG-S,n,MIN;")); 
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.70.1.1.3.0"), new OctetString("1307510455"));
                    col.Add(new Oid("1.3.6.1.2.1.1.6.0"), new OctetString("SRA MLM ALARM"));
                    col.Add(new Oid("1.3.6.1.4.1.6889.2.70.2.1.1.0"), new OctetString("10.218.112.44"));
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.1"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.2"), new OctetString("1000270571"));
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.4"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.5"), new OctetString("")); 
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.6"), new OctetString(""));
                    col.Add(new Oid("1.3.6.1.4.1.37137.2.70.1.1.7"), new OctetString("The alarm forwarded to the SRA from a product via SNMP with a trap used for acknowledgment."));
                    break;

                default:
                    mLogger.Info("CreateVbColl: Don't recognize alarm type " + alarmType);

                    col.Add(new Oid("1.3.6.1.2.1.1.1.0"), new OctetString("Test string"));
                    col.Add(new Oid("1.3.6.1.2.1.1.2.0"), new Oid("1.3.6.1.9.1.1.0"));
                    col.Add(new Oid("1.3.6.1.2.1.1.3.0"), new TimeTicks(2324));
                    col.Add(new Oid("1.3.6.1.2.1.1.4.0"), new OctetString("CreateVbColl unknown alarm type"));

                    break;
            }

            return col;
        }

        #endregion
    }
}
